# Rebulid author: Liu Pei <liupei200546@163.com>  2021/10/10

from typing import Literal, Optional, Union

import numpy as np
from matplotlib.pyplot import Axes
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from scipy.optimize import curve_fit

from qos_tools.analyzer.tools import (get_background_elimination,
                                      get_convolve_arg, get_expsinesquared_arg,
                                      get_interpolation_peaks, get_peaks,
                                      get_positive_fft_abs)
from qos_tools.experiment.lib import generate_RB_cycle
from qos_tools.fit import FitClass, T1, Rabi, Ramsey, Ramsey_withBeat, SpinEcho
from waveforms.math.fit import get_threshold_info
from waveforms.visualization import plotALLXY, plotEllipse

__all__ = [
    'analyzer_S21_abs_min', 'analyzer_S21_abs_min_multiple',
    'analyzer_spectrum',
    'analyzer_Rabi_fit', 'analyzer_Rabi_arg',
    'analyzer_Ramsey', 'analyzer_SpinEcho', 'analyzer_T1',
    'analyzer_Readout2', 'analyzer_Scatter2',
    'analyzer_AllXY', 'analyzer_AllXYtwo',
    'analyzer_RB', 'analyzer_XEB',
    'analyzer_ReadoutDelay', 'analyzer_RTO',
    'check_min', 'check_max', 'analyzer_nop',
]


def plot_population(x, ax: Optional[Axes] = None, population: Optional[np.ndarray] = None):

    if ax is not None and population is not None:
        for ipop in [1, 2, 4]:
            ax.plot(x, population[ipop], '.--', lw=1, ms=2, label=f'{ipop}')
        ax.plot(x, population[0], '--', lw=1, ms=2, label=f'drop')


def analyzer_S21_abs_min(x: np.ndarray, y: np.ndarray,
                         ker: np.ndarray = np.array([0.25, 0.5, 0.25]),
                         alpha: float = 1, ext: str = 'min', interp_points_times: float = 3,
                         ax: Optional[Axes] = None, kind: str = 'cubic', **kw) -> tuple[bool, tuple[float, ...]]:
    """
    Find the frequency according to the min abs of S21. Use `get_convolve_arg`.

    Args:
        x (np.array): frequency.
        y (np.array): S21 amplitude.
        ker (np.ndarray, optional): convolution kernel function. Defaults to np.array([0.25, 0.5, 0.25]).
        alpha (float, optional): the ratio of the difference between extreme value and mean value to standard deviation. Defaults to 1.
        ext (str, optional): extreme value type, maximum, minimum, or both. Defaults to 'min'.
        interp_points_times (float, optional): The ratio of interpolation points to original data points. Defaults to 3.
        ax (Axes, optional): the plot ax. Defaults to None.
        kind (str, optional): interpolation function type, which is used in `numpy.interp1d`. Defaults to 'cubic'.

    Returns:
        tuple[bool, tuple[float, ...]]: flag for success, x corresponding to the minimum y.
    """

    # flag, freq = get_convolve_arg(
    #     x=x, y=y, ker=ker, alpha=alpha, ext=ext, interp_points_times=interp_points_times,
    #     ax=ax, kind=kind)
    flag, freq = get_expsinesquared_arg(
        x=x, y=y, ker=ker, alpha=np.inf, ext=ext, interp_points_times=interp_points_times,
        ax=ax, **kw)

    return flag, (freq,)


def analyzer_S21_abs_min_multiple(x: np.ndarray, y: np.ndarray,
                                  ext: Literal['max', 'min', 'both'] = 'min',
                                  kind: str = 'cubic', interp_points_times: float = 3,
                                  savgol_length_times: Optional[float] = None, savgol_order: int = 1,
                                  height: Optional[float] = None, peak_height_times: float = 0.5,
                                  distance_times: float = 0.1,
                                  ax: Optional[Axes] = None, **kw) -> tuple[bool, tuple[tuple[float, ...], ...]]:
    """
    Find multiple frequencies according to the min abs of S21. Use `get_interpolation_peaks`.

    Args:
        x (np.array): frequency.
        y (np.array): S21 amplitude.
        ker (np.ndarray, optional): convolution kernel function. Defaults to np.array([0.25, 0.5, 0.25]).
        alpha (float, optional): the ratio of the difference between extreme value and mean value to standard deviation. Defaults to 1.
        ext (str, optional): extreme value type, maximum, minimum, or both. Defaults to 'min'.
        interp_points_times (float, optional): The ratio of interpolation points to original data points. Defaults to 3.
        ax (Axes, optional): the plot ax. Defaults to None.
        kind (str, optional): interpolation function type, which is used in `numpy.interp1d`. Defaults to 'cubic'.

    Returns:
        tuple[bool, tuple[float, ...]]: flag for success, x corresponding to the minimum y.
    """

    freq = get_interpolation_peaks(
        x=x, y=y, ext=ext, kind=kind, interp_points_times=interp_points_times,
        savgol_length_times=savgol_length_times, savgol_order=savgol_order, height=height,
        peak_height_times=peak_height_times, distance_times=distance_times, ax=ax)

    return len(freq) > 0, (freq,)


def analyzer_spectrum(x: np.ndarray, y: np.ndarray,
                      ext: Literal['max', 'min', 'both'] = 'max',
                      kind: str = 'cubic', interp_points_times: float = 3,
                      savgol_length_times: Optional[float] = None, savgol_order: int = 1,
                      height: Optional[float] = None, peak_height_times: float = 0.5,
                      distance_times: float = 0.1,
                      ax: Optional[Axes] = None, **kw) -> tuple[bool, tuple[float, ...]]:
    """
    Find spectrum position. Use `get_interpolation_peaks`.

    Args:
        x (np.array): frequency.
        y (np.array): S21 amplitude.
        ker (np.ndarray, optional): convolution kernel function. Defaults to np.array([0.25, 0.5, 0.25]).
        alpha (float, optional): the ratio of the difference between extreme value and mean value to standard deviation. Defaults to 1.
        ext (str, optional): extreme value type, maximum, minimum, or both. Defaults to 'max'.
        interp_points_times (float, optional): The ratio of interpolation points to original data points. Defaults to 3.
        ax (Axes, optional): the plot ax. Defaults to None.
        kind (str, optional): interpolation function type, which is used in `numpy.interp1d`. Defaults to 'cubic'.

    Returns:
        tuple[bool, tuple[float, ...]]: flag for success, x corresponding to the minimum y.
    """

    freq = get_interpolation_peaks(
        x=x, y=y, ext=ext, kind=kind, interp_points_times=interp_points_times,
        savgol_length_times=savgol_length_times, savgol_order=savgol_order, height=height,
        peak_height_times=peak_height_times, distance_times=distance_times, ax=ax)

    return len(freq) > 0, (0 if len(freq) == 0 else freq[-1],)


def analyzer_Rabi_fit(x: np.ndarray, y: np.ndarray,
                      tol: float = 0.8, goal_ans: list[str] = ['pplen', ],
                      height: Optional[float] = None, peak_height_times: float = 0.5,
                      distance_times: float = 0.1,
                      fit_times_tr: Optional[float] = None, fit_phi: float = np.pi,
                      bound_fit: float = 0.2, population: Optional[np.ndarray] = None,
                      ax: Optional[Axes] = None, arg_use: bool = False, **kw) -> Union[FitClass,
                                                                                       tuple[bool, tuple[float, ...]]]:
    """
    Return the pi pulse fitting data, time Rabi or power Rabi, using fitCalss `Rabi`.

    Args:
        x (np.ndarray): time or scale.
        y (np.ndarray): readout signal.
        tol (float, optional): tolerance of fitting goodness. Defaults to 0.8.
        goal_ans (list[str], optional): the target return value of this fit. Defaults to ['pplen', ].
        height (Optional[float], optional): the height in finding peak. Defaults to None.
        peak_height_times (float, optional): ratio of the lowerbound height of a peak to the maximum difference. Defaults to 0.5.
        distance_times (float, optional): ratio of the distance to the number of points between the peaks. Defaults to 0.1.
        fit_times_tr (Optional[float], optional): the initial value of the fitting parameter, with bound (0, np.inf). Defaults to None, i.e. np.inf.
        fit_phi (float, optional): the initial value of the fitting parameter, with bound (-np.pi*2, np.pi*2). Defaults to np.pi.
        bound_fit (float, optional): Fitting parameter boundaries. `A` is (A*(1-bound_fit), A*(1+bound_fit)), `B` is (-A*(2+bound_fit), A*(2+bound_fit)), and `Omega` is (fft_ans*(1-bound_fit), fft_ans*(1+bound_fit)). Defaults to 0.2.
        ax (Optional[Axes], optional): the plot ax. Defaults to None.
        arg_use (bool, optional): whether this fit is used in find arg. Defaults to False.

    Raises:
        ValueError: Searching peaks or maximum value in Rabi analyzer failed.

    Returns:
        Union[FitClass, tuple[bool, tuple[float, ...]]]: return fitclass or flag of success and return ans controlled by `goal_ans`
    """

    ydata = get_background_elimination(y)
    freq, amplitude = get_positive_fft_abs(x, ydata)

    freq_peak_position, _ = get_peaks(
        x=amplitude, height=height, peak_height_times=peak_height_times, distance_times=distance_times)

    p0 = {
        'A': (np.max(y)-np.min(y))/2,
        'B': np.mean(y),
        'phi': fit_phi,
    }

    if len(freq_peak_position) == 0:
        flag, pp = get_convolve_arg(x=freq, y=amplitude, ext='max', alpha=0.1)
        if flag:
            p0['Omega'] = pp
        else:
            raise ValueError(
                f'Searching peaks or maximum value in Rabi analyzer failed.')
    else:
        peak_value, amp = freq[freq_peak_position], amplitude[freq_peak_position]
        peak = np.average(peak_value, weights=amp)
        p0['Omega'] = peak

    l_bound = [p0['A']*(1-bound_fit), -p0['A'] *
               (2+bound_fit), -np.pi*2, p0['Omega']*(1-bound_fit)]
    h_bound = [p0['A']*(1+bound_fit),  p0['A']*(2+bound_fit),
               np.pi*2, p0['Omega']*(1+bound_fit)]

    fixed = {}
    if fit_times_tr is not None:
        p0['Tr'] = fit_times_tr
        l_bound.append(0)
        h_bound.append(np.inf)
    else:
        fixed['Tr'] = np.inf

    bounds = [l_bound, h_bound]

    flag = False
    try:
        RabiFit = Rabi((x, y), fixed=fixed, p0=p0)
        r2 = RabiFit.score()
        flag = r2 > tol
    except:
        pass

    if ax is not None:
        ax.plot(x, y, '.-', lw=1, ms=2)
        if population is not None:
            for ipop in range(population.shape[0]):
                ax.plot(x, population[ipop], '.-', lw=1, ms=2, label=f'{ipop}')
        if flag:
            xNew = np.linspace(np.min(x), np.max(x), 1001)
            ax.plot(xNew, RabiFit.func(xNew), '-',
                    lw=1, ms=2, label=f'$R^2={r2}$')

    if arg_use:
        return None if not flag else RabiFit

    return flag, tuple([0 if not flag else RabiFit.params[goal] for goal in goal_ans])


def analyzer_Rabi_arg(x: np.ndarray, y: np.ndarray,
                      tol: float = 0.8, ax: Optional[Axes] = None,
                      interp_points_times: float = 3, decay_times: float = 0,
                      ext: Literal['max', 'min', 'both'] = 'max',
                      kind: str = 'cubic',
                      savgol_length_times: Optional[float] = None, savgol_order: int = 1,
                      height: Optional[float] = None, peak_height_times: float = 0.5,
                      distance_times: float = 0.1, population: Optional[np.ndarray] = None,
                      alpha_times: float = 0.1, **kw) -> tuple[bool, tuple[float, ...]]:
    """
    Return the pi pulse fitting data, time Rabi or power Rabi, by finding the argmax.

    Args:
        x (np.ndarray): time or scale.
        y (np.ndarray): readout signal.
        tol (float, optional): tolerance of fitting goodness. Defaults to 0.8.
        ax (Optional[Axes], optional): the plot ax. Defaults to None.
        interp_points_times (float, optional): The ratio of interpolation points to original data points. Defaults to 3.
        decay_times (float, optional): a parameter that represents the decay times. The periodic oscillation is multiplied by the previous attenuation to avoid the selection of the second or third period in the multi-period peak search, at the cost of sacrificing some accuracy.a parameters to avoid. Defaults to 0.
        ext (Literal['max', 'min', 'both'], optional): extreme value type, maximum, minimum, or both. Defaults to 'max'.
        kind (str, optional): interpolation function type, which is used in `numpy.interp1d`. Defaults to 'cubic'.
        savgol_times (Optional[float], optional): the parameters for savgol filter. Defaults to None.
        savgol_order (int, optional): the parameters for savgol filter. Defaults to 1.
        height (Optional[float], optional): the height in finding peak. Defaults to None.
        peak_height_times (float, optional): ratio of the lowerbound height of a peak to the maximum difference. Defaults to 0.5.
        distance_times (float, optional): ratio of the distance to the number of points between the peaks. Defaults to 0.1.

    Returns:
        tuple[bool, tuple[float, ...]]: return fitclass or flag of success and pi pulse paramters
    """

    plot_population(x=x, ax=ax, population=population)

    if population is not None:
        y = population[2]-population[1]

    # RabiFit = analyzer_Rabi_fit(x=x, y=y, arg_use=True, tol=tol, ax=ax)

    # if RabiFit is not None:
    #     xNew = np.linspace(np.min(x), np.max(
    #         x), interp_points_times*x.shape[0]+1)
    #     decay_rate = np.exp(-xNew*decay_times/np.max(x))
    #     flag, peak = get_convolve_arg(x=xNew, y=RabiFit.func(
    #         xNew)*decay_rate, ext=ext, alpha=0.2, ax=ax)

    #     return flag, (peak, )

    # decay_rate = np.exp(-x*decay_times/np.max(x))

    # peaks = get_interpolation_peaks(x=x, y=y*decay_rate, ext=ext, kind=kind,
    #                                 interp_points_times=interp_points_times, savgol_length_times=savgol_length_times, savgol_order=savgol_order,
    #                                 height=height, peak_height_times=peak_height_times,
    #                                 distance_times=distance_times, ax=ax)

    # if len(peaks) == 0:
    #     flag, peak = get_convolve_arg(
    #         x, y*decay_rate, ext=ext, ax=ax, alpha=0.1)
    #     if flag:
    #         peaks = [peak]

    flag, peak = get_expsinesquared_arg(
        x=x, y=y, ext=ext, ax=ax, alpha=alpha_times*(np.max(y)-np.min(y)), **kw)
    return flag, (peak, )


def analyzer_Ramsey(x: np.ndarray, y: np.ndarray,
                    tol: float = 0.8, ax: Optional[Axes] = None, goal_ans: list[str] = ['Delta'],
                    height: Optional[float] = None, peak_height_times: float = 0.3,
                    distance_times: float = 0.01,
                    fit_t1: Optional[float] = None, fit_tphi: Optional[float] = None,
                    bound_fit: float = 0.2, fit_delta2: Optional[float] = None,
                    fit_phi: float = 0, fit_phi2: float = np.pi/2, population: Optional[np.ndarray] = None, **kw) -> tuple[bool, tuple[float, ...]]:
    """
    Analyze Ramsey result, using fit class `Ramsey` first or `Ramsey_withBeat` if `Ramsey` failed.

    Args:
        x (np.ndarray): delay time.
        y (np.ndarray): readout signal.
        tol (float, optional): tolerance of fitting goodness. Defaults to 0.8.
        ax (Optional[Axes], optional): the plot ax. Defaults to None.
        goal_ans (list[str], optional): the target return value of this fit. Defaults to ['Delta'].
        height (Optional[float], optional): the height in finding peak. Defaults to None.
        peak_height_times (float, optional): ratio of the lowerbound height of a peak to the maximum difference. Defaults to 0.5.
        distance_times (float, optional): ratio of the distance to the number of points between the peaks. Defaults to 0.01.
        fit_t1 (Optional[float], optional): fitting T1, with bound (0, np.inf). Defaults to None, i.e. np.inf.
        fit_tphi (Optional[float], optional): fitting Tphi, with bound (0, np.inf). Defaults to None, i.e. np.inf.
        bound_fit (float, optional): Fitting parameter boundaries. `A` is (A*(1-bound_fit), A*(1+bound_fit)), `B` is (-A*(2+bound_fit), A*(2+bound_fit)), and `Delta` and `Delta2` is (fft_ans*(1-bound_fit), fft_ans*(1+bound_fit)). Defaults to 0.2.
        fit_delta2 (Optional[float], optional): fitting the second frequency of oscillation. Defaults to none, i.e. try to analyze from fft spectrum.
        fit_phi (float, optional): fitting phi with bound (-2*np.pi, 2*np.pi). This is sensitive in fitting. Defaults to 0.
        fit_phi2 (float, optional): fitting phi2 with bound (-2*np.pi, 2*np.pi). Defaults to np.pi/2.

    Returns:
        tuple[bool, tuple[float, ...]]: flag of success and return ans controlled by `goal_ans`
    """

    plot_population(x=x, ax=ax, population=population)

    ydata = get_background_elimination(y)
    freq, amp = get_positive_fft_abs(x, ydata)
    peak_position, _ = get_peaks(x=amp, height=height, peak_height_times=peak_height_times,
                                 distance_times=distance_times)
    p0 = {
        'A': (np.max(ydata)-np.min(ydata))/2,
        'B': 0,
        'phi': fit_phi,
    }

    l_bound = [p0['A']*(1-bound_fit), -p0['A']*(2+bound_fit), -np.pi*2]
    h_bound = [p0['A']*(1+bound_fit),  p0['A']*(2+bound_fit),  np.pi*2]

    fixed = dict(T1=np.inf if fit_t1 is None else fit_t1)

    if fit_tphi is not None:
        l_bound.append(0)
        h_bound.append(np.inf)
        p0['Tphi'] = fit_tphi
    else:
        fixed['Tphi'] = np.inf

    if ax is not None:
        ax.plot(x, y, '.-', lw=1, ms=2)

    if len(peak_position):

        flag = False
        p0['Delta'] = freq[peak_position[0]]
        l_bound.append(p0['Delta']*(1-bound_fit))
        h_bound.append(p0['Delta']*(1+bound_fit))

        try:
            RamseyFit = Ramsey((x, ydata), fixed=fixed,
                               p0=p0)
            flag = RamseyFit.score() > tol
        except:
            pass

        if flag:
            if ax is not None:
                ax.plot(x, ydata, '.', lw=1, ms=2)
                ax.plot(x, RamseyFit.func(x), '-', lw=1, ms=2)
            return flag, tuple([RamseyFit.params[goal] for goal in goal_ans])
        else:
            if len(peak_position) > 1:
                p0['Delta'] = np.average(
                    freq[peak_position], weights=amp[peak_position])
                p0['Delta2'] = np.abs(
                    freq[peak_position[1]]-freq[peak_position[0]])
            else:
                delta = p0['Delta']
                delta2 = fit_delta2 if fit_delta2 is not None else get_convolve_arg(x=amp[peak_position[0]:],
                                                                                    y=amp[peak_position[0]:], alpha=0.1, ext='max')
                p0['Delta'], p0['Delta2'] = delta+delta2, np.abs(delta-delta2)

            l_bound[-1], h_bound[-1] = p0['Delta'] * \
                (1-bound_fit), p0['Delta']*(1+bound_fit)
            p0['phi2'] = fit_phi2
            l_bound.append(-np.pi*2)
            h_bound.append(np.pi*2)
            l_bound.append(p0['Delta2']*(1-bound_fit))
            h_bound.append(p0['Delta2']*(1+bound_fit))

            try:
                RamseyFit = Ramsey_withBeat(
                    (x, ydata), fixed=fixed, p0=p0)
                flag = RamseyFit.score() > tol
            except:
                pass

            if flag:
                if ax is not None:
                    ax.plot(x, ydata, '.', lw=1, ms=2)
                    ax.plot(x, RamseyFit.func(x), '-', lw=1, ms=2)
                return flag, tuple([RamseyFit.params[goal] for goal in goal_ans])

    return False, (0, )


def analyzer_SpinEcho(x: np.ndarray, y: np.ndarray,
                      tol: float = 0.8, ax: Optional[Axes] = None, goal_ans: list[str] = ['Delta'],
                      height: Optional[float] = None, peak_height_times: float = 0.3,
                      distance_times: float = 0.01,
                      fit_t1: Optional[float] = None, fit_tphi: Optional[float] = None, fit_phi: float = 0,
                      bound_fit: float = 0.2, population: Optional[np.ndarray] = None, **kw) -> tuple[bool, tuple[float, ...]]:
    """
    Analyze spin echo/CP/CPMG result, using fit class `SpinEcho`.

    Args:
        x (np.ndarray): delay time.
        y (np.ndarray): readout signal.
        tol (float, optional): tolerance of fitting goodness. Defaults to 0.8.
        ax (Optional[Axes], optional): the plot ax. Defaults to None.
        goal_ans (list[str], optional): the target return value of this fit. Defaults to ['Delta'].
        height (Optional[float], optional): the height in finding peak. Defaults to None.
        peak_height_times (float, optional): ratio of the lowerbound height of a peak to the maximum difference. Defaults to 0.5.
        distance_times (float, optional): ratio of the distance to the number of points between the peaks. Defaults to 0.01.
        fit_t1 (Optional[float], optional): fitting T1, with bound (0, np.inf). Defaults to None, i.e. np.inf.
        fit_tphi (Optional[float], optional): fitting Tphi, with bound (0, np.inf). Defaults to None, i.e. np.inf.
        bound_fit (float, optional): Fitting parameter boundaries. `A` is (A*(1-bound_fit), A*(1+bound_fit)), `B` is (-A*(2+bound_fit), A*(2+bound_fit)), and `Delta` is (fft_ans*(1-bound_fit), fft_ans*(1+bound_fit)). Defaults to 0.2.
        fit_phi (float, optional): fitting phi with bound (-2*np.pi, 2*np.pi). This is sensitive in fitting. Defaults to 0.

    Returns:
        tuple[bool, tuple[float, ...]]: flag of success and return ans controlled by `goal_ans`
    """

    if population is not None:
        for ipop in range(population.shape[0]):
            ax.plot(x, population[ipop], '.-', lw=1, ms=2, label=f'{ipop}')

    ydata = get_background_elimination(y)
    freq, amp = get_positive_fft_abs(x, ydata)
    peak_position, _ = get_peaks(
        x=amp, height=height, peak_height_times=peak_height_times, distance_times=distance_times)
    p0 = {
        'A': (np.max(ydata)-np.min(ydata))/2,
        'B': 0,
        'phi': fit_phi,
    }

    l_bound = [p0['A']*(1-bound_fit), -p0['A']*(2+bound_fit), -np.pi*2]
    h_bound = [p0['A']*(1+bound_fit),  p0['A']*(2+bound_fit),  np.pi*2]

    fixed = dict(T1=np.inf if fit_t1 is None else fit_t1)

    if fit_tphi is not None:
        l_bound.append(0)
        h_bound.append(np.inf)
        p0['Tphi'] = fit_tphi
    else:
        fixed['Tphi'] = np.inf

    if ax is not None:
        ax.plot(x, y, '.-', lw=1, ms=2)

    if len(peak_position):

        flag = False
        p0['Delta'] = freq[peak_position[0]]
        try:
            SpinEchoFit = SpinEcho(
                (x, ydata), fixed=fixed, p0=p0)
            flag = SpinEchoFit.score() > tol
        except:
            pass

        if flag:
            if ax is not None:
                ax.plot(x, ydata, '.', lw=1, ms=2)
                ax.plot(x, SpinEchoFit.func(x), '-', lw=1, ms=2)

            return flag, tuple([SpinEchoFit.params[goal] for goal in goal_ans])

    return False, (0, )


def analyzer_T1(x: np.ndarray, y: np.ndarray,
                tol: float = 0.8, goal_ans: list[str] = ['T1'], ax: Optional[Axes] = None,
                fit_t1: float = 100e-6, bound_fit: float = 0.2, population: Optional[np.ndarray] = None, **kw) -> tuple[bool, tuple[float, ...]]:
    """
    Analyze T1 result, using fit class `T1`.

    Args:
        x (np.ndarray): wait time.
        y (np.ndarray): readout signal.
        tol (float, optional): tolerance of fitting goodness. Defaults to 0.8.
        goal_ans (list[str], optional): the target return value of this fit. Defaults to ['T1'].
        ax (Optional[Axes], optional): the plot ax. Defaults to None.
        fit_t1 (float, optional): fitting T1, with bound (0, np.inf). Defaults to 100e-6.
        bound_fit (float, optional): fitting parameter boundaries. `A` is (A*(1-bound_fit), A*(1+bound_fit)), `B` is (-A*(1+bound_fit), A*(1+bound_fit)). Defaults to 0.2.

    Returns:
        tuple[bool, tuple[float, ...]]: flag of success and return ans controlled by `goal_ans`.
    """

    if population is not None:
        for ipop in range(population.shape[0]):
            ax.plot(x, population[ipop], '.-', lw=1, ms=2, label=f'{ipop}')

    p0 = {
        'A': np.max(y)-np.min(y),
        'B': np.mean(y),
        'T1': fit_t1,
    }

    l_bound = [-p0['A']*(1+bound_fit), -p0['B']*(1+bound_fit), 0]
    h_bound = [p0['A']*(1+bound_fit),  p0['B']*(1+bound_fit), np.inf]

    if ax is not None:
        ax.plot(x, y, '.-', lw=1, ms=2)

    flag = False
    try:
        T1Fit = T1((x, y), fixed={}, p0=p0)
        flag = T1Fit.score() > tol
    except:
        pass

    if flag:
        if ax is not None:
            ax.plot(x, y, '.', lw=1, ms=2)
            ax.plot(x, T1Fit.func(x), '-', lw=1, ms=2)

        return flag, tuple([T1Fit.params[goal] for goal in goal_ans])

    return False, (0, )


def analyzer_Readout2(x: np.ndarray, S0: np.ndarray, S1: np.ndarray,
                      ax: Optional[Axes] = None, ax1: Optional[Axes] = None,
                      alpha: float = 0.5, beta: float = 0.5,
                      alpha_times: float = 0.1,
                      r: float = 2, **kw) -> tuple[bool, tuple[float, ...]]:
    """
    Analyzer 0 state (ground state) and 1 state (the first excited state) result, using support vector machine.

    Args:
        x (np.ndarray): independent variable array. 
        S0 (np.ndarray): 0 state readout.
        S1 (np.ndarray): 1 state readout.
        ax (Optional[Axes], optional): the main plot ax. Defaults to None.
        ax1 (Optional[Axes], optional): the secondary plot ax. Defaults to None.
        alpha (float, optional): coefficient of distance between two states in the cost function. Defaults to 0.5.
        beta (float, optional): coefficient of visibility between two states in the cost function. Defaults to 0.5.
        r (float, optional): the radius multiple of two ellipses in a scatter plot. Defaults to 2.

    Returns:
        tuple[bool, tuple[float, ...]]: the flag of success and the parameter corresponds to the maximum separate position.
    """

    dis = []
    vis = []
    for i in range(x.shape[0]):
        info = get_threshold_info(S0[i], S1[i])
        dis.append(np.abs(info['center'][0]-info['center'][1]))
        vis.append(info['visibility'][0])
    dis = np.asarray(dis)
    dis = dis/np.max(dis)
    vis = np.asarray(vis)

    cost = alpha*(dis)**2+beta*vis**2

    # flag, arg = get_convolve_arg(
    #     x=range(x.shape[0]), y=cost, ext='max', alpha=0.5)
    flag, arg = get_expsinesquared_arg(interp_points_times=1, x=np.arange(x.shape[0]), y=cost, ext='max', ax=ax, alpha=alpha_times*(np.max(cost)-np.min(cost)), **kw)
    arg = round(arg)
    if ax is not None:
        if flag:
            if ax1 is not None:
                ax1.plot(np.real(S0[arg]), np.imag(
                    S0[arg]), 'C0.', lw=1, ms=2, alpha=0.5)
                ax1.plot(np.real(S1[arg]), np.imag(
                    S1[arg]), 'C1.', lw=1, ms=2, alpha=0.5)
                plotEllipse(c0=info['center'][0], a=r*info['std']
                            [0], b=r*info['std'][1], phi=info['phi'], ax=ax1)
                plotEllipse(c0=info['center'][1], a=r*info['std']
                            [2], b=r*info['std'][3], phi=info['phi'], ax=ax1)

    return flag, (x[arg], )


def analyzer_Scatter2(S0: np.ndarray, S1: np.ndarray,
                      ax: Optional[Axes] = None, tol: float = 0.8,
                      hot_thresh: int = 10000, r: float = 2, **kw) -> tuple[bool, tuple[float, ...]]:
    """
    Analyzer the readout information in 0 and 1 states, using support vector machine.

    Args:
        S0 (np.ndarray): 0 state readout.
        S1 (np.ndarray): 1 state readout.
        ax (Optional[Axes], optional): the plot ax. Defaults to None.
        tol (float, optional): the lower bound of visibility. Defaults to 0.8.
        hot_thresh (int, optional): used in plot process. Defaults to 10000.
        r (float, optional): the radius multiple of two ellipses in a scatter plot. Defaults to 2.

    Returns:
        tuple[bool, tuple[float, ...]]: flag of success and readout info
    """

    info = get_threshold_info(S0, S1)

    if ax is not None:

        ax.set_xticks([])
        ax.set_yticks([])
        for s in ax.spines.values():
            s.set_visible(False)

        re0, re1 = info['signal']

        ax3 = inset_axes(ax, width='65%', height='65%', loc='lower left')

        ax3.hist(re0, bins=50, alpha=0.5)
        ax3.hist(re1, bins=50, alpha=0.5)

        x, a, b, c = info['cdf']
        ax1 = ax3.twinx()
        ax1.plot(x, a)
        ax1.plot(x, b)
        ax1.plot(x, c)
        ax1.set_ylim(0, 1.1)
        ax3.set_title(f"{info['visibility'][0]:.4f}")

        ax2 = inset_axes(ax, width='30%', height='30%', loc='upper right')

        if (len(S0) + len(S1)) < hot_thresh:
            ax2.plot(np.real(S0), np.imag(S0), '.', ms=2, alpha=0.5)
            ax2.plot(np.real(S1), np.imag(S1), '.', ms=2, alpha=0.5)
        else:
            _, *bins = np.histogram2d(np.real(np.hstack([S0, S1])),
                                      np.imag(np.hstack([S0, S1])),
                                      bins=50)

            H0, *_ = np.histogram2d(np.real(S0),
                                    np.imag(S0),
                                    bins=bins,
                                    density=True)
            H1, *_ = np.histogram2d(np.real(S1),
                                    np.imag(S1),
                                    bins=bins,
                                    density=True)
            vlim = max(np.max(np.abs(H0)), np.max(np.abs(H1)))

            ax2.imshow(
                H1.T - H0.T, alpha=(np.fmax(H0.T, H1.T) / vlim).clip(0, 1), interpolation='nearest', origin='lower', cmap='coolwarm', vmin=-vlim,
                vmax=vlim, extent=(bins[0][0], bins[0][-1], bins[1][0], bins[1][-1]))

        for s in ax2.spines.values():
            s.set_visible(False)

        plotEllipse(c0=info['center'][0]*np.exp(1j*info['phi']), a=r*info['std'][0],
                    b=r*info['std'][1], phi=info['phi'], ax=ax2)
        plotEllipse(c0=info['center'][1]*np.exp(1j*info['phi']), a=r*info['std'][2],
                    b=r*info['std'][3], phi=info['phi'], ax=ax2)

        im0, im1 = info['idle']
        lim = min(im0.min(), im1.min()), max(im0.max(), im1.max())
        t = (np.linspace(lim[0], lim[1], 3) + 1j *
             info['threshold']) * np.exp(-1j * info['phi'])
        ax2.plot(t.imag, t.real, 'k--')

        ax2.plot(np.real(info['center'][0]*np.exp(1j*info['phi'])), np.imag(
            info['center'][0]*np.exp(1j*info['phi'])), 'o', color='C3')
        ax2.plot(np.real(info['center'][1]*np.exp(1j*info['phi'])), np.imag(
            info['center'][1]*np.exp(1j*info['phi'])), 'o', color='C4')

        ax2.axis('equal')
        ax2.set_xticks([])
        ax2.set_yticks([])

    return info['visibility'][0] > tol, (info['threshold'], info['phi'])


def analyzer_AllXY(y: np.ndarray,
                   ax: Optional[Axes] = None, tol: float = 0.98, population: Optional[np.ndarray] = None, 
                   ave_goal: Optional[np.ndarray] = None, **kw) -> tuple[bool, tuple[float, ...]]:
    """
    Plot AllXY result and try to analyze it. 'error = sum(std)+sum(|value - goal|)/4.'

    Args:
        y (np.ndarray): AllXY result.
        ax (Optional[Axes], optional): the plot ax. Defaults to None.
        tol (float, optional): the upper bound on the error. Defaults to 0.98.

    Returns:
        tuple[bool, tuple[float, ...]]: flag of success and the error.
    """

    plot_population(x=range(21), ax=ax, population=population)

    pos = [np.arange(0, 5), np.arange(5, 17), np.arange(17, 21)]

    data_std = []
    data_ave = []
    for i, item in enumerate(pos):
        data_std.append(np.std(y[item]))
        data_ave.append(np.mean(y[item]))
    data_ave = np.asarray(data_ave)
    data_std = np.asarray(data_std)
    print(data_ave)
    ave_goal = data_ave if ave_goal is None else ave_goal

    err = np.sum(data_std)+np.sum(np.abs(data_ave-ave_goal))/4

    if ax is not None:
        plotALLXY(y, ax=ax)
        ax.legend(title=f'{err}')

    return err < tol, (err, )


def analyzer_AllXYtwo(x: np.ndarray, y1: np.ndarray, y2: np.ndarray,
                      ax: Optional[Axes] = None,
                      interp_points_times: float = 3, polyfit_order: int = 8, population: Optional[np.ndarray] = None, **kw) -> tuple[bool, tuple[float, ...]]:
    """
    In the DRAG parameters calibration, find the intersection of two terms in AllXY series.

    Args:
        x (np.ndarray): independent variable array.
        y1 (np.ndarray): result of one item in AllXY series.
        y2 (np.ndarray): result of another item in AllXY series.
        ax (Optional[Axes], optional): the plot ax. Defaults to None.
        interp_points_times (float, optional): The ratio of interpolation points to original data points. Defaults to 3.
        polyfit_order (int, optional): use `np.polyfit` to smooth the result, this is the order paramster. Defaults to 8.

    Returns:
        tuple[bool, tuple[float, ...]]: flag of success and the best parameter.
    """

    # plot_population(x=x, ax=ax, population=population)

    fit1 = np.polyfit(x, y1, polyfit_order)
    fit2 = np.polyfit(x, y2, polyfit_order)
    func1 = np.poly1d(fit1)
    func2 = np.poly1d(fit2)

    xx = np.linspace(np.min(x), np.max(x), round(
        interp_points_times*x.shape[0])+1)
    flag = (func1(np.min(x))-func2(np.min(x))) * \
        (func1(np.max(x))-func2(np.max(x))) < 0

    if ax is not None:
        ax.plot(x, y1, '.-', lw=1, ms=2)
        ax.plot(x, y2, '.-', lw=1, ms=2)

        ax.plot(xx, func1(xx), '-', lw=1, ms=2)
        ax.plot(xx, func2(xx), '-', lw=1, ms=2)

        ax.axvline(x=xx[np.argmin(np.abs(func1(xx)-func2(xx)))],
                   c='k', ls='-', lw=1)

    return flag, (xx[np.argmin(np.abs(func1(xx)-func2(xx)))], )


def rb_error(p, pc, d, ave=None, p_ref=None, pc_ref=None):
    if p_ref is None:
        return (1-p)*(1-d)/ave, np.sqrt(pc)*(1-d)/ave
    else:
        return (1-p/p_ref)*(1-d), np.abs(p/p_ref)/2*np.sqrt(pc/p**2+pc_ref/p_ref**2)


def analyzer_RB(y: list[np.ndarray], max_cycle: int, sweep_points: int, random_times: int,
                 interleaves: list[str] = [], ave: float = 1.5, start: int = 1, d = 1/2,
                 ax = None, ref_index: int = 1, tol: float = 0.998, share_base=False, **kw) -> tuple[bool, tuple[float, ...]]:

    def rb_curve(n, p, A, B):
        return A * p**n + B
    
    from scipy.optimize import curve_fit
    from qos_tools.experiment.libs.tools import generate_log_intlist
    x = generate_log_intlist(max_int=max_cycle, sweep_points=sweep_points, min_int=start)
    xx = np.linspace(x.min(), x.max(), 101)

    yy, ee = np.mean(y, axis=1), np.std(y, axis=1)/np.sqrt(random_times)
    
    p = []
    p_cov = []
    if share_base:
        _popt, _pcov = curve_fit(rb_curve, x, yy[-1], sigma=ee[-1], p0=[0.99, 0.5, 0.25], maxfev=100000)
        y_base = _popt[2]
        ax.axhline(y=y_base, c='k', ls='--')
    for i, interleave in enumerate(interleaves):
        if share_base:
            _popt, _pcov = curve_fit(lambda n, p, A, B=y_base: rb_curve(n, p, A, B), 
                                     x, yy[i], sigma=ee[i], p0=[0.99, 0.5], maxfev=100000)
        else:
            _popt, _pcov = curve_fit(rb_curve, x, yy[i], sigma=ee[i], p0=[0.99, 0.5, 0.25], maxfev=100000)
        p.append(_popt)
        p_cov.append(_pcov)
    from qos_tools.analyzer.tools import get_error_output
    
    ret = []
    if ax is not None:
        for i, interleave in enumerate(interleaves):
            for j in range(random_times):
                ax.plot(x, y[i, j, :], f'C{i%10}.', alpha=0.2)
                
        for i in range(len(interleaves)):
            if i<ref_index:
                epg, epg_err, epg_dig = get_error_output(*(rb_error(p=p[i][0], pc=p_cov[i][0][0], d=d, ave=ave)))
            else:
                epg, epg_err, epg_dig = get_error_output(*(rb_error(p=p[i][0], pc=p_cov[i][0][0], d=d, 
                                                                    p_ref=p[0][0], pc_ref=p_cov[0][0][0])))              
            ret.extend([p[i][0], p_cov[i][0][0], epg, epg_err, epg_dig])
            epg_print = 'EPG={x:.%df}({y})' % epg_dig
            if share_base:
                ax.plot(xx, rb_curve(xx, *p[i], y_base), f'C{i%10}-')
            else:
                ax.plot(xx, rb_curve(xx, *p[i]), f'C{i%10}-')
            ax.errorbar(x, yy[i], ee[i], fmt='x', capsize=2, c=f'C{i%10}', label=f'{interleaves[i]}: '+epg_print.format(x=epg, y=epg_err))
    return True, np.asarray(ret).reshape([-1, 5])


def analyzer_XEB(seeds: np.ndarray, cycle: np.ndarray, counts: np.ndarray,
                 ax: Optional[Axes] = None, tol: float = 0.998, population: Optional[np.ndarray] = None, **kw) -> tuple[bool, tuple[float, ...]]:

    # def rb_curve(n, p, A, B=0.25):
    #     return A*p**n+B

    # from waveforms.quantum.xeb import Fxeb2

    # ans = []
    # for i, icycle in enumerate(cycle):
    #     ans.append(Fxeb2(qubits=tuple(select5), cycle=icycle, seeds=ran, counts=count[i, :], shots=1024))
    # ans = np.asarray(ans)

    # if ax is not None:
    #     ax.plot(x, y, '.', label='data', ms=2)

    # p = 0
    # try:
    #     popt, pcov = curve_fit(rb_curve, x, y, p0=[0.99, 0.5, 0.5])
    #     p = popt[0]
    #     if ax is not None:
    #         xx = np.linspace(x.min(), x.max(), 1001)
    #         ax.plot(xx, rb_curve(xx, *popt), '-', label='fit', lw=1, ms=2)
    #         ax.legend(title=f'p={100*(popt[0]):.4f}')
    # except:
    #     pass

    p = 0
    return p > tol, (p, )


def analyzer_ReadoutDelay(x: np.ndarray, y: np.ndarray,
                          ax: Optional[Axes] = None,
                          interp_points_times: float = 10, ker: np.array = np.array([0.1, 0.2, 0.4, 0.2, 0.1]),
                          kind: str = 'cubic', population: Optional[np.ndarray] = None, alpha_times: float = 0.1, **kw) -> tuple[bool, tuple[float, ...]]:
    """
    Find the readout delay for probe line, using `get_convolve_arg`.

    Args:
        x (np.ndarray): delay time.
        y (np.ndarray): readout signal.
        ax (Optional[Axes], optional): the plot ax. Defaults to None.
        interp_points_times (float, optional): The ratio of interpolation points to original data points. Defaults to 10.
        ker (np.array, optional): convolve kernel function. Defaults to np.array([0.1, 0.2, 0.4, 0.2, 0.1]).
        kind (str, optional): interpolation function type, which is used in `numpy.interp1d`. Defaults to 'cubic'.

    Returns:
        tuple[bool, tuple[float, ...]]: [description]
    """

    # flag, delay = get_convolve_arg(
    #     x=x, y=y, ker=ker, ext='max', interp_points_times=interp_points_times, ax=ax, kind=kind)
    flag, delay = get_expsinesquared_arg(x=x, y=y, ext='max', interp_points_times=interp_points_times, ax=ax,
                                         alpha=alpha_times*(np.max(y)-np.min(y)))
    return flag, (delay, )


def check_min(x: np.ndarray, y: np.ndarray,
              ax: Optional[Axes] = None, **kw) -> tuple[bool, tuple[float, ...]]:
    """
    Quick check for min value, x's and y's first dimension is 3 and y[1] is the min.

    Args:
        x (np.array): independent variable.
        y (np.array): dependent variable.
        ax (Optional[Axes], optional): [description]. Defaults to None.

    Returns:
        tuple[bool, tuple[float, ...]]: [description]
    """

    if ax is not None:
        ax.plot(x, y, '.-', lw=1, ms=2)

    assert y.shape[0] == 3, 'check number is valid.'

    return np.all((y[0] >= y[1]) and (y[1] <= y[2])), (0,)


def check_max(x: np.ndarray, y: np.ndarray,
              ax: Optional[Axes] = None, **kw) -> tuple[bool, tuple[float, ...]]:
    """
    Quick check for max value, x's and y's first dimension is 3 and y[1] is the max.

    Args:
        x (np.array): independent variable.
        y (np.array): dependent variable.
        ax (Optional[Axes], optional): [description]. Defaults to None.

    Returns:
        tuple[bool, tuple[float, ...]]: [description]
    """

    if ax is not None:
        ax.plot(x, y, '.-', lw=1, ms=2)

    assert y.shape[0] == 3, 'check number is valid.'

    return np.all((y[0] <= y[1]) and (y[1] >= y[2])), (0,)


def analyzer_RTO(x: np.ndarray, y: np.ndarray,
                 ax: Optional[Axes] = None,
                 tol: float = 0.1e6, population: Optional[np.ndarray] = None, **kw) -> tuple[bool, tuple[float, ...]]:
    """
    Analyze the result of RTO.

    Args:
        x (np.ndarray): delay time.
        y (np.ndarray): x-axis projection and y-axis projection.
        ax (Optional[Axes], optional): the plot ax. Defaults to None.
        tol (float, optional): upper limit of frequency variation. Defaults to 0.1e6.

    Returns:
        tuple[bool, tuple[float, ...]]: flag of success and the frequency change.
    """

    if population is not None:
        for ipop in range(population.shape[0]):
            ax.plot(x, population[ipop], '.--', lw=1, ms=2, label=f'{ipop}')

    assert y.shape[0] == 2*x.shape[0], 'Invalid x and y shape'
    y = y.reshape([x.shape[0], 2])
    IQ = y[:, 0]+1j*y[:, 1]
    diff_IQ = np.diff(IQ)
    angle_IQ = np.unwrap(np.angle(diff_IQ), discont=np.pi)
    accuracy = np.mean(np.diff(angle_IQ))/(x[1]-x[0])/np.pi/2

    if ax is not None:
        ax.plot(y[:, 0]*2-1, y[:, 1]*2-1, '.-', lw=1, ms=2)
        ax.scatter(y[0, 0]*2-1, y[0, 1]*2-1)
        ax.axis('equal')

    return accuracy < tol, (accuracy,)


def analyzer_nop(
        flag: bool = True, ax: Optional[Axes] = None, **kw) -> tuple[bool, tuple[float, ...]]:
    """
    A virtual analyzer for nop scanner.

    Args:
        flag (bool, optional): always return flag. Defaults to True.
        ax (Optional[Axes], optional): the plot ax. Defaults to None.

    Returns:
        tuple[bool, tuple[float, ...]]: flag, (0, )
    """

    return flag, (0, )



# Rebuild: Liu Pei  <liupei200546@163.com>  2022/02/25

from typing import Optional

import numpy as np


def generate_spanlist_linear(center: Optional[float] = None,
                             delta: Optional[float] = None,
                             st: Optional[float] = None,
                             ed: Optional[float] = None,
                             sweep_points: int = 101,
                             **kw) -> np.ndarray:
    """
    Linear span. Two cases are allowed:
        - `center` and `delta` given, an array on (`center`-`delta`, `center`+`delta`) with sweep points equal to `sweep_points` is returned.
        - `st` for start and `ed` for end, an an array on (`st`, `ed`) with sweep points equal to `sweep_points` is returned.

    Args:
        center (Optional[float], optional): Center. Defaults to None.
        delta (Optional[float], optional): Span. Defaults to None.
        st (Optional[float], optional): Start. Defaults to None.
        ed (Optional[float], optional): End. Defaults to None.
        sweep_points (int, optional): Sweep points. Defaults to 101.

    Raises:
        ValueError: Illegal paramsters

    Returns:
        np.ndarray: 1d array of data
    """

    if center is not None and delta is not None:
        return np.linspace(center - delta, center + delta, sweep_points)
    elif st is not None and ed is not None:
        return np.linspace(st, ed, sweep_points)
    else:
        raise ValueError('The scanning interval entered is illegal')


def generate_spanlist_log(center: Optional[float] = None,
                          delta: Optional[float] = None,
                          st: Optional[float] = None,
                          ed: Optional[float] = None,
                          sweep_points: int = 101,
                          **kw) -> np.ndarray:
    """
    Log span. Three cases are allowed:
        - `center` and `delta` given, an array on (`center`-`delta`, `center`) and (`center`, `center`+`delta`) with sweep points equal to half of `sweep_points` for each is returned.
        - `st` for start and `ed` for end, an array on (`st`, `ed`) with sweep points equal to `sweep_points` is returned.
        - `st`, `ed` and `center` given, an array on (`st`, `center`) and (`center`, `ed`) with sweep points equal to half of `sweep_points` for each is returned.

    Args:
        center (Optional[float], optional): Center. Defaults to None.
        delta (Optional[float], optional): Span. Defaults to None.
        st (Optional[float], optional): Start. Defaults to None.
        ed (Optional[float], optional): End. Defaults to None.
        sweep_points (int, optional): Sweep points. Defaults to 101.

    Raises:
        ValueError: Illegal paramsters

    Returns:
        np.ndarray: 1d array of data
    """

    sweep_half = (np.logspace(0, 1, (sweep_points + 1) // 2) - 1) / 9
    if center is not None and delta is not None:
        return np.concatenate(
            (-sweep_half[::-1][:-1], sweep_half)) * delta + center
    elif st is not None and ed is not None:
        if center is None:
            center = (st + ed) / 2
            delta = (ed - st) / 2
            return np.concatenate(
                (-sweep_half[::-1][:-1], sweep_half)) * delta + center
        else:
            return np.concatenate(
                (sweep_half[::-1][:-1] * (st - center), sweep_half *
                 (ed - center))) + center
    else:
        raise ValueError('The scanning interval entered is illegal')


def generate_spanlist(mode: str = 'linear', **kw) -> np.ndarray:
    """
    Generate a span list.

    Args:
        mode (str, optional): 'linear' for linear span and 'log' for log. Defaults to 'linear'.

    Raises:
        ValueError: f'Mode {mode} is not supported.'

    Returns:
        np.ndarray: 1d array of data
    """
    if mode == 'linear':
        return generate_spanlist_linear(**kw)
    elif mode == 'log':
        return generate_spanlist_log(**kw)
    else:
        raise ValueError(f'Mode {mode} is not supported.')


def generate_log_intlist(max_int: int, sweep_points: int, min_int: float = 1) -> np.ndarray:
    """
    Generate a list of int which follows the log-uniform.

    Args:
        max_int (int): Maximum integer.
        sweep_points (int): Expected point number, but the returned on may be less than it because of unique method. 
        min_int (float, optional): Minimum integer. Defaults to 1.

    Returns:
        np.ndarray: 1d array of int
    """

    return np.unique(np.round(np.logspace(np.log10(min_int*1.0), np.log10(max_int*1.0), sweep_points))).astype(int)



# Author: lizhiyuan  <lizhiyuan@baqis.ac.cn>  2021/01/29
# Add score: Liu Pei  <liupei200546@163.com>  2022/03/18
# rebuild: lizhiyuan  <lizhiyuan@baqis.ac.cn>  2022/04/15


import logging
from inspect import signature

import matplotlib.pyplot as plt
import numpy as np
from scipy import interpolate
from scipy.optimize import leastsq, least_squares
from sklearn.metrics import r2_score

# from qos_tools.analyzer.statistics import get_R2
# from qos_tools.fit.GUI import guiqwtfit
log = logging.getLogger(__name__)


class FitClass(object):
    """Fit class, based on scipy.optimiz.least_squares/leastsq """

    def __init__(self, data, fitfunc, fixed=None, p0=None, indep_var=0,
                 optimize_mode='leastsq', **kw):
        '''
        data: tuple (x,y),数据
        fitfunc: 拟合函数
        fixed: dict, 确定的参数
        p0: dict/list/tuple/array, 初始拟合参数,
            dict,可以只给部分参数的初值,其他初值默认为1;
            list/tuple/array,所有拟合参数的初值
        indep_var: str/int, 自变量,
            如果是int,则表示第n个参数为自变量,
            如果是str,则表示自变量的名字
        optimize_mode: least_squares/leastsq, 两版优化模式
        '''
        x, y = data
        self.x = np.asarray(x)
        self.y = np.asarray(y)

        self.fitfunc = fitfunc
        self.fixed = {} if fixed is None else fixed

        all_keys = list(signature(fitfunc).parameters.keys())
        self.indep_var = all_keys[indep_var] if isinstance(
            indep_var, int) else indep_var

        all_keys.remove(self.indep_var)
        for k in self.fixed.keys():
            all_keys.remove(k)
        self.fit_keys = all_keys

        p0 = {} if p0 is None else p0
        if isinstance(p0, dict):
            self.p0 = [p0.get(k, 1) for k in self.fit_keys]
        else:
            self.p0 = p0
        self._plotscript = None

        if optimize_mode == 'least_squares':
            self.perform_fit_by_least_squares(**kw)
        elif optimize_mode == 'leastsq':
            self.perform_fit_by_leastsq(**kw)
        else:
            raise Exception(f"optimize mode '{optimize_mode}' not support!")

    def gen_errFunc(self):
        def errFunc(params, x, y):
            # print(params)
            params_dict = dict()
            fit_params_dict = dict(zip(self.fit_keys, params))
            params_dict.update(fit_params_dict)
            params_dict.update(self.fixed)
            params_dict.update({self.indep_var: x})
            # print(params_dict)
            e = y - self.fitfunc(**params_dict)
            return np.abs(e)
        return errFunc

    def perform_fit_by_least_squares(self, **least_squares_kw):
        errFunc = self.gen_errFunc()
        res = least_squares(errFunc, self.p0, args=(
            self.x, self.y), **least_squares_kw)
        popt = res.x
        try:
            # covariance matrix when jac not degenerate
            J = res.jac
            pcov = np.linalg.inv(J.T.dot(J))
            # standard deviation errors on the parameters
            fit_error = dict(zip(self.fit_keys, np.sqrt(np.diag(pcov))))
        except Exception as e:
            fit_error = dict(zip(self.fit_keys, [np.nan]*len(self.fit_keys)))
            log.warning(f'failed to calculate pcov:{repr(e)}')

        self.res = res
        self.params = {}
        self.fited = dict(zip(self.fit_keys, popt))
        self.params.update(self.fited)
        self.params.update(self.fixed)
        self.error = {}
        self.error.update(fit_error)
        fix_error = dict(zip(self.fixed.keys(), [0]*len(self.fixed.keys())))
        self.error.update(fix_error)

    def perform_fit_by_leastsq(self, **leastsq_kw):
        errFunc = self.gen_errFunc()
        res = leastsq(errFunc, self.p0, args=(
            self.x, self.y), full_output=1, **leastsq_kw)
        popt, pcov, infodict, errmsg, ier = res
        if pcov is None:
            log.warning(f'ier: {ier}, msg: {errmsg}',)
            fit_error = dict(zip(self.fit_keys, [np.nan]*len(self.fit_keys)))
        else:
            # 根据curve_fit源代码重新计算协方差矩阵
            cost = np.sum(infodict['fvec'] ** 2)
            ysize = len(infodict['fvec'])
            s_sq = cost / (ysize - popt.size)
            pcov = pcov * s_sq
            # standard deviation errors on the parameters
            fit_error = dict(zip(self.fit_keys, np.sqrt(np.diag(pcov))))

        self.res = res
        self.params = {}
        self.fited = dict(zip(self.fit_keys, popt))
        self.params.update(self.fited)
        self.params.update(self.fixed)
        self.error = {}
        self.error.update(fit_error)
        fix_error = dict(zip(self.fixed.keys(), [0]*len(self.fixed.keys())))
        self.error.update(fix_error)

    def func(self, t):
        '''拟合后的函数'''
        indep_var_kw = {self.indep_var: t}
        return self.fitfunc(**indep_var_kw, **self.fited, **self.fixed)

    def gen_smooth_data(self, smooth_level=10):
        """产生一组拟合后的采样更密的画图数据
        smooth_level: deflault 10,插值的倍率,重新对x轴数据插值使画出的拟合曲线更平滑
        """
        t, y = self.x, self.y
        index_sorted = np.argsort(t)
        t, y = t[index_sorted], y[index_sorted]

        size = len(t)
        t_func = interpolate.interp1d(np.arange(size), t, kind='linear')
        _t = t_func(np.linspace(0, size-1, int((size-1)*smooth_level+1)))
        _y = self.func(_t)
        return _t, _y

    def plot(self, fmt='r-', plotscript=None):
        '''
        画图
            Parameters:
                fmt: plot curve format
                plotscript: 额外的画图脚本函数,接收self和ax两个参数
        '''
        t, y = self.x, self.y
        _t, _y = self.gen_smooth_data()

        ax = plt.gca()
        ax.scatter(t, y,)
        ax.plot(_t, _y, fmt)

        if plotscript is not None:
            plotscript(self, ax=ax)
        elif self._plotscript is not None:
            self._plotscript(self, ax=ax)
        else:
            pass

    def plot2(self, plot_fmt='-', modes=None, plot_option=None):
        '''
        画图2
            Parameters:
                modes: real/imag/amp(abs)/phase(angle) 等的列表或字符串
                plot_fmt: plot format
                plot_option: plot  keywords
        '''
        if modes is None:
            modes = ['real']
        elif isinstance(modes, str):
            modes = [modes]

        ax = plt.gca()
        t, y = self.x, self.y
        _t, _y = self.gen_smooth_data()

        plot_kw = {}
        if plot_option is not None:
            plot_kw.update(plot_option)

        if 'real' in modes:
            ax.scatter(t, y.real)
            ax.plot(_t, _y.real, plot_fmt, label='real', **plot_kw)
        if 'imag' in modes:
            ax.scatter(t, y.imag)
            ax.plot(_t, _y.imag, plot_fmt, label='imag', **plot_kw)
        if 'amp' in modes or 'abs' in modes:
            ax.scatter(t, np.abs(y))
            ax.plot(_t, np.abs(_y), plot_fmt, label='amp', **plot_kw)
        if 'phase' in modes or 'angle' in modes:
            ax.scatter(t, np.angle(y))
            ax.plot(_t, np.angle(_y), plot_fmt, label='phase', **plot_kw)

    def score(self, sample_weight=None) -> float:
        y_pred = self.func(self.x)
        return r2_score(self.y, y_pred, sample_weight=sample_weight)



# Author: lizhiyuan  <lizhiyuan@baqis.ac.cn>  2021/01/01
# Priority algorithm support by Liupei

import numpy as np


def extract_align_info(stdgatemap):
    '''提取对齐位置信息字典'''
    align_info = {}
    for qubit, gatelist in stdgatemap.items():
        for i, stdindex in enumerate(gatelist):
            operator, qubits, params = stdindex
            if operator.count(':'):
                if operator not in align_info:
                    align_info[operator] = {}
                align_info[operator][qubit] = i
    return align_info


def extract_gateMap_priority(stdgatemap, align_info):
    '''计算gateMap_priority'''
    gateMap_priority = {
        qubit: np.array(range(len(gatelist)))
        for qubit, gatelist in stdgatemap.items()
    }

    for qubit in stdgatemap.keys():
        flt = filter((lambda v, q=qubit: q in v.keys()),
                     align_info.values())
        sort = sorted(flt, key=lambda v, q=qubit: v[q])
        for align_dct in sort:
            priority_max = max(gateMap_priority[qubit][idx]
                               for qubit, idx in align_dct.items())
            for qubit, idx in align_dct.items():
                p_add = priority_max - gateMap_priority[qubit][idx]
                gateMap_priority[qubit][idx:] += p_add
    return gateMap_priority


def extract_align_priority_dict(gateMap_priority, align_info):
    '''提取对齐标记的优先度列表'''
    align_priority_dict = {}
    for align_tag, dct in align_info.items():
        align_qubits = list(dct.keys())
        qubit, gindex = list(dct.items())[0]
        align_priority = gateMap_priority[qubit][gindex]
        align_priority_dict[align_priority] = align_qubits
    return align_priority_dict


def extract_priority_info(gateMap_priority):
    '''按照优先度分类，提取排版顺序信息'''
    priority_info = {}
    for qubit, prioritylist in gateMap_priority.items():
        for i, priority in enumerate(prioritylist):
            if priority not in priority_info:
                priority_info[priority] = []
            priority_info[priority].append((qubit, i))
    return priority_info


def extract_gate_info(stdgatemap, priority_info):
    cmds = []
    for priority in sorted(priority_info.keys()):
        _cmds = []
        _barr_qubit_list = []
        for q, i in priority_info[priority]:
            stdindex = stdgatemap[q][i]
            operator, qubits, params = stdindex
            if len(qubits) > 1:
                # 校验多比特门的对齐规范
                _stdindex = stdgatemap[q][i-1]
                _operator, _qubits, _params = _stdindex
                assert _operator.count(
                    ':'), f"Qubit '{q}' {i+1}-th gate '{stdindex}' miss alignment mark !!!"
                assert q in qubits, f"Qubit '{q}' not support '{stdindex}'!"
            elif len(qubits) == 1:
                assert qubits[0] == q, f"Qubit '{q}' {i+1}-th gate '{stdindex}' error!"
            else:
                qubits = (q,)

            if operator.count(':'):
                _barr_qubit_list.extend(qubits)
            else:
                _cmds.append((operator, qubits, params))

        # 取集合，消除重复的多比特门操作
        _cmds_set = set(_cmds)
        # 校验 ############################
        qubits_op = []
        for operator, qubits, params in _cmds_set:
            qubits_op.extend(qubits)
        assert len(_cmds) == len(qubits_op) == len(
            set(qubits_op)), f'part of multi-qubit gate missing!'
        # ############################
        cmds.extend(_cmds_set)

        if _barr_qubit_list:
            barr_cmd = ('Barrier', tuple(_barr_qubit_list), tuple())
            cmds.append(barr_cmd)
    return cmds



# Author: Liu Pei  <liupei200546@163.com>  2022/03/10

from typing import Optional

import numpy as np
from matplotlib.axes import Axes
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import (RBF, ConstantKernel,
                                              ExpSineSquared, WhiteKernel)


def plot_gpr(x: np.ndarray,
             y: np.ndarray,
             gpr: GaussianProcessRegressor,
             x_new: Optional[np.ndarray] = None,
             y_new: Optional[np.ndarray] = None,
             y_std: Optional[np.ndarray] = None,
             interp_points_times: float = 10,
             lw: float = 1,
             ms: float = 2,
             ax: Optional[Axes] = None,
             **kw) -> None:

    if x_new is None or y_new is None or y_std is None:
        x_new = np.linspace(np.min(np.real(x)), np.max(np.real(x)),
                            round(interp_points_times * x.shape[0]) + 1)
        y_new, y_std = gpr.predict(x_new.reshape(-1, 1), return_std=True)
        y_new, y_std = y_new.reshape([
            y_new.shape[0],
        ]), y_std.reshape([
            y_std.shape[0],
        ])

    if ax is not None:
        ax.fill_between(x_new, y_new - y_std, y_new + y_std, alpha=0.2)
        ax.plot(x, y, '.', lw=lw, ms=ms)
        ax.plot(x_new, y_new, '-', lw=lw, ms=ms)


def preprocess_1d(x: np.ndarray,
                  mode: str = 'MinMaxScaler',
                  **kw) -> np.ndarray:
    """
    Preprocess for 1d array scaler. `sklearn.preprocessing` are used here.
    Support mode: 'MinMaxScaler', 'MaxAbsScaler', 'RobustScaler', 'StandardScaler'

    Args:
        y (np.ndarray): 1d array.
        mode (str, optional): `sklearn.preprocessing` calss name. Defaults to 'MinMaxScaler'.

    Returns:
        _type_: _description_
    """

    import sklearn.preprocessing as preprocess_lib
    return getattr(preprocess_lib,
                   mode)(**kw).fit_transform(X=x.reshape([-1, 1]))[:, 0]


def exp_sin_kernel(x: np.ndarray, y: np.ndarray,
                   **kw) -> tuple[GaussianProcessRegressor, float]:
    fit_kernel = ConstantKernel()*ExpSineSquared()*RBF() + \
        ConstantKernel()+WhiteKernel()
    gpr = GaussianProcessRegressor(kernel=fit_kernel,
                                   normalize_y=kw.get('normalize_y', False))
    return gpr.fit(x.reshape(-1, 1), y.reshape(-1, 1)+5e-5), 1/gpr.kernel_.get_params()['k1__k1__k1__k2__periodicity']



# Author: lizhiyuan  <lizhiyuan@baqis.ac.cn>  2021/01/01

import numpy as np


def _T1(t, A, B, T1):
    y = A*np.exp(-t/T1)+B
    return y


def _Rabi(t, A, B, phi, Omega, Tr):
    y = A*np.exp(-t/Tr)*np.cos(2*np.pi*Omega*t+phi)+B
    return y


def _Ramsey(t, A, B, phi, T1, Tphi, Delta):
    y = A*np.exp(-t/2/T1-np.square(t/Tphi))*np.cos(2*np.pi*Delta*t+phi)+B
    return y


def _Ramsey_withBeat(t, A, B, phi, T1, Tphi, Delta, phi2, Delta2):
    y = A*np.exp(-t/2/T1-np.square(t/Tphi))*np.cos(
        2*np.pi*Delta*t+phi)*np.cos(2*np.pi*Delta2*t+phi2)+B
    return y


def _RB(t, A, B, p):
    y = A*p**t+B
    return y


def Linear(t, k, b):
    y = k * t + b
    return y


def Sin(t, A, B, f, phi):
    y = A*np.sin(2*np.pi*f*t+phi)+B
    return y


def Cauchy(t, A, t0, FWHM):
    y = A*FWHM/((t-t0)**2+FWHM**2)/np.pi
    return y


def Gaussian(x, A, miu, sigma):
    return A*np.exp(-((x-miu)/sigma)**2/2)


def F_ge(I, F_ge_max, Period, I_SS, d):
    '''
    基态和第一激发态能级间距与磁通偏置电流的关系
    F_ge是指基态和第一激发态能级间距

    I：偏置电流大小
    F_ge_max: 最高点频率，
    Period: 对应一个磁通量子的电流周期，
    I_SS: 频率最高点对应的电流大小，
    d: SQUID双结的不对称度参数，完全对称为0，完全不对称为1,
    '''
    phi = np.pi * (I - I_SS) / Period
    y = F_ge_max * (np.cos(phi)**2 + d * d * np.sin(phi)**2)**0.25
    return y


def F_cav(I, F_cav_bare, g_qr, F_ge_max, Period, I_SS, d):
    '''
    色散区域的谐振腔频率与磁通偏置电流的关系
    F_cav是谐振腔频率

    I：偏置电流大小
    F_cav_bare: 裸腔频率，
    g_qr: 谐振腔与量子比特的耦合常量，
    F_ge_max: 最高点频率，
    Period: 对应一个磁通量子的电流周期，
    I_SS: 频率最高点对应的电流大小，
    d: SQUID双结的不对称度参数，完全对称为0，完全不对称为1,
    '''
    F_ge_i = F_ge(I, F_ge_max, Period, I_SS, d)
    term1 = (F_cav_bare + F_ge_i)/2
    term2 = np.sqrt(g_qr * g_qr + (F_ge_i - F_cav_bare)**2 / 4)
    a = np.where(F_cav_bare > F_ge_i, 1, -1)
    y = term1 + term2*a
    return y



# Author: Liu Pei  <liupei200546@163.com>  2022/03/24

from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import scipy.linalg as sl
from matplotlib.axes import Axes
from qos_tools.analyzer.tools import get_positive_fft_abs
from scipy.optimize import Bounds, curve_fit, minimize
from sklearn.metrics import r2_score


def _act2msr(sigma: np.ndarray, p1: float = 1, p2: float = 1) -> np.ndarray:
    """
    Add a classical measurement noise to the result of state tomography. 

    Args:
        sigma (np.ndarray): Actual tomography result.
        p1 (float, optional): P(measure = 0|prepare = 0). Defaults to 1.
        p2 (float, optional): P(measure = 1|prepare = 1).. Defaults to 1.

    Returns:
        np.ndarray: Result of measure.
    """
    return (p1+p2-1)*sigma+p1-p2


def _msr2act(sigma: np.ndarray, p1: float = 1, p2: float = 1) -> np.ndarray:
    """
    Calculate the actual result of tomography from a classical measurement noise. 

    Args:
        sigma (np.ndarray): Tomography result.
        p1 (float, optional): P(measure = 0|prepare = 0). Defaults to 1.
        p2 (float, optional): P(measure = 1|prepare = 1). Defaults to 1.

    Returns:
        np.ndarray: Actual result of tomography.
    """
    return (sigma-p1+p2)/(p1+p2-1)


def _sphere2coordinates(x: np.ndarray) -> np.ndarray:
    """
    Transformation from spherical coordinate to rectangular coordinate system.

    Args:
        x (np.ndarray): Spherical coordinate.

    Returns:
        np.ndarray: rectangular coordinate.
    """
    return np.array([x[0]*np.sin(np.pi*x[1])*np.cos(2*np.pi*x[2]), x[0]*np.sin(np.pi*x[1])*np.sin(2*np.pi*x[2]), x[0]*np.cos(np.pi*x[1])])


def _theory_coeff(theta: float, phi: float) -> np.ndarray:
    """
    Coefficient of solution.

    Args:
        theta (float): Oscillation parameter.
        phi (float): Oscillation parameter.

    Returns:
        np.ndarray: Coefficient of evolution curve.
    """

    ox, oy, oz = np.sin(np.pi*theta)*np.cos(2*np.pi*phi), np.sin(np.pi *
                                                                 theta)*np.sin(2*np.pi*phi), np.cos(np.pi*theta)
    con_coeff = np.array([[ox*ox, ox*oy, ox*oz],
                          [oy*ox, oy*oy, oy*oz],
                          [oz*ox, oz*oy, oz*oz]])
    cos_coeff = np.array([[1-ox*ox,   -ox*oy,   -ox*oz],
                          [-oy*ox, 1-oy*oy,   -oy*oz],
                          [-oz*ox,   -oz*oy, 1-oz*oz]])
    sin_coeff = np.array([[0, -oz,  oy],
                          [oz,   0, -ox],
                          [-oy,  ox,   0]])

    return np.array([con_coeff, cos_coeff, sin_coeff])


def _cr_cos_fit(msr: np.ndarray, t: np.ndarray, r2: float = 0.5, ax: Optional[Axes] = None, r: float = 0) -> tuple[float, float]:
    """
    Fitting evolution curve with sinusoidal oscillation.

    Args:
        msr (np.ndarray): Measure result.
        t (np.ndarray): Evolution time.
        r2 (float, optional): R2 of fitting. Defaults to 0.5.
        ax (Optional[Axes], optional): Plot axes. Defaults to None.
        r (float, optional): Oscillation period. Defaults to 0.

    Returns:
        tuple[float, float]: Oscillation parameter theta and phi
    """

    def f_cos(x, a, b, c, d):
        return a*np.cos(2*np.pi*b*x+c)+d

    ret = []
    for i, s in enumerate(msr):
        try:
            popt, pcov = curve_fit(f_cos, t, s, p0=[0.5, r, 0, 0], maxfev=1000000)
            if r2_score(s, f_cos(t, *popt)) > r2:
                ret.append(popt[0])
                if ax is not None:
                    ax.plot(t, f_cos(t, *popt), '--', lw=0.5, ms=2)
            else:
                ret.append((np.max(s)-np.min(s))/2)
        except:
            ret.append((np.max(s)-np.min(s))/2)

    ret = np.asarray(ret)
    sin_theta, tan_phi = np.sqrt(2)*np.abs(ret[2]), np.sqrt(
        np.abs((ret[2]**2+ret[0]**2-ret[1]**2)/(ret[2]**2-ret[0]**2+ret[1]**2)))

    return np.arcsin(sin_theta if sin_theta < 1 else 1), np.arctan(tan_phi)


def theory_evolution(r: float, theta: float, phi: float, sx: float, sy: float, sz: float, p1: float = 1, p2: float = 1, t: Optional[np.ndarray] = None, t0: Optional[float] = None) -> np.ndarray:
    """
    Theoretical evolution curve.

    Args:
        r (float): Oscillation period.
        theta (float): Oscillation parameter.
        phi (float): Oscillation parameter.
        sx (float): Result of tomography of x axis at the initial time.
        sy (float): Result of tomography of y axis at the initial time.
        sz (float): Result of tomography of z axis at the initial time.
        p1 (float, optional): P(measure = 0|prepare = 0). Defaults to 1.
        p2 (float, optional): P(measure = 1|prepare = 1). Defaults to 1.
        t (Optional[np.ndarray], optional): Evolution time. Defaults to None.
        t0 (Optional[float], optional): Initial time. Defaults to None.

    Returns:
        np.ndarray: measure result of theoretical evolution curve.
    """

    coeff = _theory_coeff(theta=theta, phi=phi)
    sigma = _msr2act(np.array([sx, sy, sz]), p1=p1, p2=p2)

    cal = (coeff@sigma).T@np.array([np.ones_like(t),
                                    np.cos(2*np.pi*r*(t-t0)), np.sin(2*np.pi*r*(t-t0))])

    return _act2msr(cal, p1=p1, p2=p2)


def _delta(x: np.ndarray, y: np.ndarray) -> float:
    """
    Calculate the period of evolution.

    Args:
        x (np.ndarray): Time.
        y (np.ndarray): Signal.

    Returns:
        float: Most likely oscillation period.
    """
    freq, amp = get_positive_fft_abs(x, y)
    return float(freq[np.argmax(amp)])


def _minimize_cr(msr: np.ndarray, t: np.ndarray, t0: float, s0: np.ndarray, p1: float = 1, p2: float = 1, random_times: int = 20, r2: float = 0.8, tol: float = 1e-10, ax: Optional[Axes] = None,
                 r2_fit: float = 0.8, fft_pos: int = 2, delta: float = None, **kw) -> tuple[np.ndarray, float]:
    """
    Search evolution parameters.

    Args:
        msr (np.ndarray): Measure result.
        t (np.ndarray): Evolution time.
        t0 (float): Initial time.
        s0 (np.ndarray): Result of tomography at the initial time.
        p1 (float, optional): P(measure = 0|prepare = 0). Defaults to 1.
        p2 (float, optional): P(measure = 1|prepare = 1). Defaults to 1.
        random_times (int, optional): Search times, random evolution parameters are used since the 8th try. Defaults to 20.
        r2 (float, optional): R2 of evolution search. Defaults to 0.8.
        tol (float, optional): Parameters for `minimize` method. Defaults to 1e-10.
        ax (Optional[Axes], optional): Plot axes. Defaults to None.
        r2_fit (float, optional): R2 of fitting. Defaults to 0.8.

    Returns:
        tuple[np.ndarray, float]: Evolution parameters and R2
    """

    def _minimize_cr_eps(x: list, msr: np.ndarray, t: np.ndarray, t0: float):
        return sl.norm(msr-theory_evolution(*x, t, t0))

    _r = _delta(x=t, y=msr[fft_pos]) if delta is None else delta
    _theta, _phi = _cr_cos_fit(msr=msr, t=t, r2=r2_fit, r=_r)

    x0 = [_r, _theta, _phi, *s0, p1, p2]
    random_test = [(i, j) for i in np.arange(0, 1, 0.5)
                   for j in np.arange(0, 1, 0.25)]
    random_test.extend([(np.random.random(), np.random.random())
                       for _ in range(8, random_times)])

    for i, pp in enumerate(random_test):
        x0[1], x0[2] = np.fmod(_theta+pp[0], 1), np.fmod(_phi+pp[1], 1)
        try:
            xx = minimize(_minimize_cr_eps, x0=x0, args=(msr, t, t0), tol=tol,
                        bounds=Bounds((_r*0.8, -0.01, -0.01, -1.01, -1.01, -1.01, 0.49, 0.49),
                                        (_r*1.2,  1.01,  1.01,  1.01,  1.01,  1.01, 1.01, 1.01)))
            fit = theory_evolution(*xx.x, t=t, t0=t0)
            _score = min([r2_score(s, fit[i]) for i, s in enumerate(msr)])
            if _score > r2:
                if ax is not None:
                    ax.plot(t, msr.T, '.', lw=1, ms=2)
                    ax.plot(t, fit.T, '-', lw=1, ms=2)
                return np.array(xx.x[:3]), _score
        except:
            pass
    if ax is not None:
        ax.plot(t, msr.T, '.-', lw=1, ms=2)
    return np.array([np.nan, np.nan, np.nan]), np.nan


def analyzer_cr_tomo_1s(x: np.ndarray, y: np.ndarray, t0_pos: float = 0.5, plot_figure: bool = True, sphere: bool = False, **kw) -> np.ndarray:
    """
    Analyzer for CR Hamiltonian tomography with a single point.

    Args:
        x (np.ndarray): Evolution time.
        y (np.ndarray): Measure data.
        t0_pos (float, optional): Initial time position. Defaults to 0.5.
        plot_figure (bool, optional): Control for plot fitting curve. Defaults to True.

    Returns:
        np.ndarray: CR Coefficient and R2.
    """

    assert y.shape[0] == x.shape[0] and y.shape[1] == 6, 'Invalid data shape.'

    if plot_figure:
        fig, ax = plt.subplots(1, 2, figsize=[9, 2])
    else:
        ax = [None, None]

    t0_cnt = round(x.shape[0]*t0_pos)

    ans0, r2_0 = _minimize_cr(msr=y[:, :3].T, t=x, t0=x[t0_cnt], s0=y[t0_cnt, :3], ax=ax[0], **kw)
    ans1, r2_1 = _minimize_cr(msr=y[:, 3:].T, t=x, t0=x[t0_cnt], s0=y[t0_cnt, 3:], ax=ax[1], **kw)

    if plot_figure:
        plt.tight_layout()
        plt.show()

    if not sphere:
        ans0 = _sphere2coordinates(x=ans0)
        ans1 = _sphere2coordinates(x=ans1)
        ans0, ans1 = (ans0+ans1)/2, (ans0-ans1)/2

    return np.array([*ans0, *ans1, r2_0, r2_1])


def analyzer_cr_tomo_1d(x: np.ndarray, y: np.ndarray, t: np.ndarray, ax: Optional[Axes] = None, with_r2: bool = True, **kw) -> np.ndarray:
    """
    Analyzer for CR Hamiltonian tomography as one parameter changing.

    Args:
        x (np.ndarray): parameters.
        y (np.ndarray): Measure data.
        t (np.ndarray): Evolution time.
        ax (Optional[Axes], optional): Plot axes. Defaults to None.
        with_r2 (bool, optional): Plot R2 data or not. Defaults to True.

    Returns:
        np.ndarray: CR Coefficient and R2.
    """

    if len(t.shape)<2:
        _, tt = np.meshgrid(x, t)
        tt = tt.T
    else:
        tt = t

    y = y.reshape([-1, t.shape[1], 6])
    assert x.shape[0] == y.shape[0], 'Invalid data shape'

    coeff_list = np.array([analyzer_cr_tomo_1s(x=tt[i], y=yy, **kw) for i, yy in enumerate(y)])

    if ax is not None:
        label_list = ['$R_0$', '$\\theta_0$', '$\phi_0$', '$R_1$', '$\\theta_1$', '$\phi_1$'] if kw.get(
            'sphere', False) else ['IX', 'IY', 'IZ', 'ZX', 'ZY', 'ZZ']
        for i in range(6):
            ax.plot(x, coeff_list[:, i], '.-', lw=1, ms=2, label=label_list[i])
        if with_r2:
            ax.plot(x, coeff_list[:, 6], ':', lw=1, ms=2, label='$R^2_0$')
            ax.plot(x, coeff_list[:, 7], ':', lw=1, ms=2, label='$R^2_1$')
        if not kw.get('sphere', False):
            ax.plot(x, np.sqrt(
                coeff_list[:, 0]**2+coeff_list[:, 1]**2), '.--', lw=1, ms=2, label='IXY')
            ax.plot(x, np.sqrt(
                coeff_list[:, 3]**2+coeff_list[:, 4]**2), '.--', lw=1, ms=2, label='ZXY')

    return coeff_list



# Author: lizhiyuan  <lizhiyuan@baqis.ac.cn>  2021/01/01

import numpy as np

__all__ = [
    'is_equal', 'assert_equal',
]


def is_equal(v1, v2):
    '''深入地比较v1,v2是否相等'''
    try:
        assert_equal(v1, v2)
        return True
    except AssertionError:
        return False


def assert_equal(v1, v2):
    '''断言v1,v2相等'''
    try:
        if isinstance(v1, np.ndarray) or isinstance(v1, np.ndarray):
            assert np.all(v1 == v2)
        else:
            assert v1 == v2
    except ValueError:  # ValueError: The truth value of an array with more than one element is ambiguous. Use a.any() or a.all()
        if isinstance(v1, dict) and isinstance(v2, dict):
            assert not set(v1.keys()) ^ set(v2.keys()), 'keys not equal !'
            for k in v1.keys():
                _v1, _v2 = v1[k], v2[k]
                assert_equal(_v1, _v2)
        elif isinstance(v1, (list, tuple, set)) and isinstance(v2, (list, tuple, set)):
            for _v1, _v2 in zip(v1, v2):
                assert_equal(_v1, _v2)
        else:
            assert False



# Author: Liu Pei  <liupei200546@163.com>  2021/10/13

from typing import Any, Literal, Optional

import skopt.space.space as space
from pydantic import BaseModel
from scipy.optimize import OptimizeResult
from skopt import Optimizer

from .model import FeedAgent


class ParaMode(BaseModel):
    pat: Optional[Literal['mul', 'add']] = None
    mode: Optional[Literal['Real', 'Integer']] = None


class BayesOpt(FeedAgent):
    """
    Agend for Bayes optmizer.

    Variable in the form:
    {
        'pattern': (Literal['mul', 'add']): base+bound or base*bound.
        'low' (float): lower bound.
        'high' (float): upper bound.
        'type' (Literal['Real', 'Integer']): parameter type.
    }

    Attribute:
        opt (Optimizer): the `skopt.Optimizer` class
    """

    config = {
        'transform': 'normalize',
        'prior': 'uniform',
        'base_estimator': 'GP',
        'n_initial_points': 10,
        'initial_point_generator': 'hammersly',
    }

    def __init__(self, variable: list[dict], base: list, config: dict = {}, **kw):
        """
        Create a new opt.

        Args:
            variable (list[dict]): variable list.
            base (list): initial value.
            config (dict, optional): optimizer configuration parameters. Defaults to {}.
        """

        self.config.update(config)

        self.variable = []

        for i, item in enumerate(variable):
            self.variable.append(self._add_variable(base=base[i], **item))

        self.opt = Optimizer(self.variable,
                             base_estimator=self.config['base_estimator'],
                             n_initial_points=self.config['n_initial_points'],
                             initial_point_generator=self.config['initial_point_generator'])

    def _add_variable(self, base: Any, pattern: str = 'mul', low: Any = 0.5, high: Any = 1.5, type: str = 'Real', **kw):

        params = ParaMode(pat=pattern, mode=type)
        pattern, type = params.pat, params.mode

        if pattern in ['mul']:
            low, high = base*low, base*high
        else:
            low, high = base+low, base+high

        return getattr(space, type)(low=low, high=high, prior=self.config['prior'], transform=self.config['transform'])

    def tell(self, x, y, fit=True) -> OptimizeResult:
        """
        Feed data.

        Args:
            x ([type]): variable value.
            y ([type]): cost variable
            fit (bool, optional): fitting model. Defaults to True.

        Returns:
            OptimizeResult: current ans
        """
        return self.opt.tell(x, y, fit=fit)

    def ask(self, n_points=None, strategy="cl_min"):
        """
        Ask the next point to estimate.

        Args:
            n_points ([type], optional): n points. Defaults to None.
            strategy (str, optional): Defaults to "cl_min".
        """
        return self.opt.ask(n_points=n_points, strategy=strategy)

    def check(self, tol: float) -> bool:
        """
        Check whether the current result is less than the threshold.

        Args:
            tol (float): [description]

        Returns:
            bool: [description]
        """
        result = self.result()
        return result.fun < tol

    def result(self) -> OptimizeResult:
        """
        Get result.

        Returns:
            OptimizeResult: current result.
        """
        return self.opt.get_result()



# Author: lizhiyuan  <lizhiyuan@baqis.ac.cn>  2021/01/01

import asyncio
import functools
from collections.abc import Awaitable
from qurpc.client import ZMQRPCCallable


__all__ = [
    'asyncmethod', 'ensure_awaitable', 'seq_exe', 'sync_exe',
]


async def asyncmethod(loop, func, *args, **kw):
    if asyncio.iscoroutinefunction(func):
        result = await func(*args, **kw)
    elif isinstance(func, ZMQRPCCallable):
        result = func(*args, **kw)
        if isinstance(result, Awaitable):
            result = await result
    else:
        result = await loop.run_in_executor(
            None, functools.partial(func, *args, **kw))
        if isinstance(result, Awaitable):
            result = await result
    return result


async def ensure_awaitable(func, *args, **kw):
    if asyncio.iscoroutinefunction(func):
        result = await func(*args, **kw)
    else:
        result = func(*args, **kw)
        if isinstance(result, Awaitable):
            result = await result
    return result


async def seq_exe(taskdict):
    '''顺序执行一个任务列表'''
    res = {}
    for k, task in taskdict.items():
        r = await task
        res.update({k: r})
    return res


async def sync_exe(taskdict):
    '''同步执行一个任务列表'''
    rl = await asyncio.gather(*taskdict.values())
    res = dict(zip(taskdict.keys(), rl))
    return res



# Author: Liu Pei  <liupei200546@163.com>  2022/03/15

from typing import Optional

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from networkx import Graph


class Mapping(Graph):

    _update_flag = False
    _qubit_neighbors = {}
    _direct_coupler_count = 0

    cmap = 'nipy_spectral'
    allowed_type = ['Q', 'C', 'D', 'M', 'J']
    type_color = {'Q': 0, 'C': 1, 'D': 2, 'M': 3, 'J': 4}
    place_holder = '*'

    def check_type(self):
        for item in self.nodes:
            assert self.nodes[item]['type'] in self.allowed_type, 'Invalid type of components.'

    def _dfs(self, node, deep: Optional[int] = None, condition: Optional[list[str]] = None):

        def _check_condition(node, condition):
            return condition == self.place_holder or self.nodes[node]['type'] == condition

        def __dfs(deep, condition, dis, stack, ret, flag):
            if dis+1 >= deep:
                ret.append(tuple(stack))
            else:
                for item in self.neighbors(stack[dis]):
                    if _check_condition(node=item, condition=condition[dis+1]) and flag[item]:
                        stack[dis+1], flag[item] = item, False
                        __dfs(deep=deep, condition=condition, dis=dis +
                              1, stack=stack, ret=ret, flag=flag)
                        flag[item] = True

        if condition is None and deep is None:
            raise KeyError('Free parameters for dfs are not supported')

        self.check_type()

        if condition is None:
            condition = [self.place_holder]*deep
        elif deep is None:
            deep = len(condition)

        ret = []

        if not _check_condition(node=node, condition=condition[0]):
            return ret

        stack, flag = [node]*deep, {item: True for item in self.nodes}
        flag[node] = False

        __dfs(deep=deep, condition=condition, dis=0,
              stack=stack, ret=ret, flag=flag)
        return ret

    def _add_components(self, components: list[str], type: str):
        assert type in self.allowed_type, 'Invalid type of components.'
        self.add_nodes_from(components, type=type)
        self._update_flag = True

    def add_qubits(self, qubits: list[str], type: str = 'Q'):
        self._add_components(qubits, type)

    def add_coupler(self, coupler: str, qubits: list[str], type: str = 'C'):
        self._add_components([coupler], type)
        self.add_edges_from([(coupler, q) for q in qubits])

    def add_direct_couple(self, qubits: list[str], type: str = 'D'):
        self._direct_coupler_count += 1
        self._add_components([f'D{self._direct_coupler_count}'], type)
        self.add_edges_from(
            [(f'D{self._direct_coupler_count}', q) for q in qubits])
        return f'D{self._direct_coupler_count}'

    def add_probe(self, probe, qubits: list[str], type: str = 'M'):
        self._add_components([probe], type)
        self.add_edges_from([(probe, q) for q in qubits])

    def add_jpa(self, jpa, probes: list[str], type: str = 'J'):
        self._add_components([jpa], type)
        self.add_edges_from([(jpa, p) for p in probes])

    def _draw_node_config(self, tag, ndim):
        ret = {'cmap': self.cmap}
        vmin, vmax = 1e50, -1e50
        if tag is None:
            ret['node_color'] = [self.type_color[self.nodes[item]['type']]
                                 for item in self.nodes]
        else:
            ret['nodelist'] = []
            ret['node_color'] = []
            ret['labels'] = {}
            for item in self.nodes:
                value = self.nodes[item].get(tag, np.nan)
                if np.isnan(value):
                    continue
                ret['node_color'].append(value)
                label_output = '{x:.%df}' % ndim
                ret['labels'][item] = label_output.format(x=value)
                ret['nodelist'].append(item)
                vmin, vmax = min(vmin, value), max(vmax, value)

        return ret, vmin if vmin<1e50 else 0, vmax if vmax>-1e50 else 1

    def draw(self, tag: Optional[str] = None, layout: str = 'spring_layout', seed=1234, colorbar: bool = True,
             font_size=8, alpha=1, font_color='w', ndim: int = 0, **kw):
        from networkx import layout as layout_lib
        if 'pos' not in kw.keys():
            kw['pos'] = getattr(layout_lib, layout)(self, seed=seed)
        update_config, vmin, vmax = self._draw_node_config(tag=tag, ndim=ndim)
        kw.update(update_config)
        nx.draw_networkx(self, with_labels=True, font_size=font_size,
                         font_color=font_color, alpha=alpha, **kw)
        if 'ax' in kw.keys():
            ax = kw['ax']
            ax.set_yticks([])
            ax.set_xticks([])
            for i in ['bottom', 'top', 'right', 'left']:
                ax.spines[i].set_visible(False)
            if isinstance(tag, str):
                ax.set_title(tag)
            if tag is not None and colorbar:
                self.colorbar(ax=ax, vmin=vmin, vmax=vmax)

    def colorbar(self, fraction: float = 0.1, shrink: float = 0.5, vmin: float = 0, vmax: float = 1, **kw):
        from matplotlib.cm import ScalarMappable
        scaler = ScalarMappable(cmap=self.cmap)
        scaler.set_clim(vmin=vmin, vmax=vmax)
        plt.colorbar(scaler, fraction=fraction, shrink=shrink, **kw)

    def component_neighbors(self, node, deep: Optional[int] = None, condition: Optional[list[str]] = None):
        return self._dfs(node, deep, condition)

    def clear(self):
        self._direct_coupler_count = 0
        super().clear()


def config2mapping(keys: list[str], query: callable) -> Mapping:
    """
    From a query method to create a `Mapping` class. 

    Args:
        keys (list[str]): keys list from config.
        query (callable): query method

    Returns:
        Mapping: Graph object.
    """
    ret = Mapping()
    qubits = [key for key in keys if key[0] == 'Q']
    ret.add_qubits(qubits=qubits)
    for item in keys:
        if item[0] == 'C':
            ret.add_coupler(item, query(f'{item}.qubits'))
        elif item[0] == 'M':
            ret.add_probe(item, query(f'{item}.qubits'))
        elif item[0] == 'J':
            ret.add_jpa(item, query(f'{item}.probe'))
        else:
            pass
    return ret


def mapping2cfg(map: Mapping, update: callable):
    """
    From a `Mapping` method to update a config

    Args:
        map (Mapping): Graph object.
        update (callable): update method of config
    """

    from itertools import combinations

    from .create import (create_coupler, create_probe, create_qubit,
                         gate_mapping_1, gate_mapping_2)

    gates = {gate: {}
             for gate in list(gate_mapping_1.keys())+list(gate_mapping_2.keys())}
    for item in map.nodes:
        if map.nodes[item]['type'] == 'Q':
            couplers, probe = [], None
            for goal in map.neighbors(item):
                if map.nodes[goal]['type'] == 'C':
                    couplers.append(goal)
                elif map.nodes[goal]['type'] == 'M':
                    probe = goal
            update(item, create_qubit(probe=probe, couplers=couplers))

            for gate in gate_mapping_1.keys():
                gates[gate][item] = gate_mapping_1[gate]()

        elif map.nodes[item]['type'] == 'C':
            qubits = []
            for goal in map.neighbors(item):
                if map.nodes[goal]['type'] == 'Q':
                    qubits.append(goal)
            update(item, create_coupler(qubits=qubits))
        elif map.nodes[item]['type'] == 'M':
            qubits = []
            for goal in map.neighbors(item):
                if map.nodes[goal]['type'] == 'Q':
                    qubits.append(goal)
            update(item, create_probe(qubits=qubits))

    for gate in gate_mapping_2.keys():
        for _type, method in gate_mapping_2[gate]:
            for item in map.nodes:
                if map.nodes[item]['type'] == _type:
                    for q1, q2 in combinations([goal for goal in map.neighbors(item) if map.nodes[goal]['type'] == 'Q'], 2):
                        gates[gate][f'{q1}_{q2}'] = method()

    update('gate', gates)
